import asyncio
import json
import logging
import sys

import unittest

from rf_client import MindMap, NodeNotFound, set_config, logger, Node, UserNotFound, User

with open('./secret.json') as f:
    config = json.load(f)

set_config(config['api_url'])


def async_test(coro):
    # https://stackoverflow.com/a/46324983
    def wrapper(*args, **kwargs):
        loop = asyncio.new_event_loop()
        return loop.run_until_complete(coro(*args, **kwargs))

    return wrapper


def debug(*args):
    logger.debug(' '.join(map(str, args)))


class TestMapLoading(unittest.TestCase):
    @async_test
    async def test_root_loading(self):
        map_id = config['map_id']
        username = config['username']
        password_hash = config['password_hash']
        async with MindMap(map_id, (username, password_hash)) as mm:
            self.assertEqual(mm.root.body.children[0].parent_node, mm.root)
            self.assertEqual(mm.root.body.children[0].parent_node, mm.root.body.children[1].parent_node)
            self.assertEqual(mm.root.get_parents(), [])

            with self.assertRaises(NodeNotFound):
                mm.get('bad_uuid')

            n = mm.root.find(None)
            self.assertIsInstance(n, Node)

            nl = mm.root.find_all(None)
            self.assertIsInstance(nl, list)

            with self.assertRaises(UserNotFound):
                mm.users.find_by_id('bad_uuid')

            with self.assertRaises(UserNotFound):
                mm.users.find_by_username('bad_username')

            current_user = mm.users.find_by_username(username)
            self.assertEqual(current_user.username, username)


if __name__ == '__main__':
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.info('Tests started')

    unittest.main()



from typing import Optional, List, Dict

from dateutil import parser

from rf_client.exceptions import NodeNotFound, RFApiError
from rf_client.utils import DictMixin


class Node(DictMixin):
    """
    Модель узла
    """
    class Properties(DictMixin):
        """
        Свойства узла. Состоят из групп
        """
        def __init__(self, properties):
            if properties is None:
                properties = {}

            self.byUser = properties.get('byUser', [])  # todo map to OrderedDict
            self.byType = properties.get('byType', {})
            self.global_ = properties.get('global', {})
            self.style = properties.get('style', {})

        def __str__(self):
            return str(dict({
                'byUser': self.byUser,
                'byType': self.byType,
                'global': self.global_,
                'style': self.style
            }))

    class Body(DictMixin):
        """ Тело узла """
        def __init__(self, id, type_id=None, properties=None, parent=None, children=None, parent_node=None, unread_comments_count=None, ids_to_nodes=None, **kwargs):

            assert ids_to_nodes is not None  # fixme

            self.id = id
            self.type_id = type_id
            self.properties = Node.Properties(properties)
            self.parent = parent

            self.unread_comments_count = unread_comments_count

            if children:
                self.children = [Node(**node_dict, parent_node=parent_node, ids_to_nodes=ids_to_nodes) for node_dict in children]  # fixme
            else:
                self.children = []

        def __repr__(self):
            return f"Body('{self.id}', ...)"

    class Meta(DictMixin):
        """ Метаданные узла """

        # fixme editable and commentable
        def __init__(self, creation_timestamp, last_modified_timestamp, last_modified_user, author, **kwargs):
            try:
                self.creation_timestamp = parser.parse(creation_timestamp) if creation_timestamp else None
                self.last_modified_timestamp = parser.parse(last_modified_timestamp) if last_modified_timestamp else None
            except ValueError:
                raise RFApiError("Wrong datetime response format")

            self.last_modified_user = last_modified_user
            self.author = author

        def __str__(self):
            return str({
                'creation_timestamp': self.creation_timestamp,
                'last_modified_timestamp': self.last_modified_timestamp,
                'last_modified_user': self.last_modified_user,
                'author': self.author
            })

    def __init__(self, id, map_id, meta, position, body, parent, originalParent, hidden, properties=None,
                 unread_comments_count=None,
                 parent_node=None,
                 ids_to_nodes=None, **kwargs):

        assert ids_to_nodes is not None  # fixme
        self.ids_to_nodes: Dict[str, Node] = ids_to_nodes
        self.ids_to_nodes[id] = self

        self.id = id
        self.map_id = map_id
        self.meta = Node.Meta(**meta)
        self.position = position
        self.properties = Node.Properties(properties)
        self.parent = parent  # это id предка, так сделано из-за формата ответа с сервера
        self.parent_node = parent_node  # это ссылка на предка Node
        self.originalParent = originalParent

        self.unread_comments_count = unread_comments_count
        self.hidden = hidden

        body_id = body['id']

        self._is_link = body_id != self.id

        if not self._is_link:
            self.body = Node.Body(**body, parent_node=self, ids_to_nodes=ids_to_nodes)  # fixme
        else:
            self.body = Node.Body(body_id, ids_to_nodes=ids_to_nodes)  # fixme

    @property
    def is_link(self):
        return self._is_link

    def __repr__(self):
        title = self.body.properties.global_.get('title', 'N/A')
        title = title[:100] + '...' if len(title) > 100 else title
        return f"Node('{self.id}', '{self.body.id}', '{title}')"

    def find(self, type_id: Optional[str]) -> Optional['Node']:
        """
        Найти первого потомка с указанным типом. Понятие "первый" означает лишь любого, кто попадется первым.
        :return None | Node
        """
        def dive(children: List[Node]):
            for node in children:
                if node.is_link:
                    continue  # fixme

                if node.body.type_id == type_id:
                    return node
                else:
                    result = dive(node.body.children)
                    if result:
                        return result

        return dive(self.body.children) if self.body.type_id is not type_id else self

    def find_by_id(self, id: str) -> Optional['Node']:
        """
        Найти первого потомка с указанным id. Понятие "первый" означает лишь любого, кто попадется первым.
        :return None | Node
        """
        try:
            target_node = self.ids_to_nodes[id]
            if self in target_node.get_parents():
                return target_node
            else:
                return None
        except KeyError:
            return None

    def find_all(self, type_id: Optional[str]) -> List['Node']:
        """
        Тоже, что и find, но не останавливаемся на первом найденном, а возвращает список потомков узла
          с определенным типом.

        :return [] | [Node, ...]
        """
        def dive(children: List[Node]):
            result = []
            for node in children:
                if node.is_link:
                    continue  # fixme в нашем дереве появляется цикл из-за ссылок!
                if node.body.type_id == type_id:
                    result.append(node)

                t = dive(node.body.children)
                result += t if t is not None else []

            return result

        t = dive(self.body.children)
        return t if t is not None else []

    def find_closest_parent(self, type_id: str) -> 'Node':
        """
        Найти ближайшего предка опредленного типа type_id узла node_id.
        """
        cur = self
        while cur.body.type_id != type_id:
            cur = cur.parent_node
            if cur is None:  # Если дошли до корня
                raise NodeNotFound()

        return cur

    def get_parents(self) -> List['Node']:
        """
        Получить список родителей узла. Порядок - от узла к корню, не включая себя.
        """
        result = []
        cur = self
        while cur.parent_node is not None:
            cur = cur.parent_node
            result.append(cur)

        return result



from random import random
import logging
import ujson
from json import JSONDecodeError
from asyncio import ensure_future

import aiohttp
from aiohttp import ClientSession, BasicAuth, ClientTimeout

from typing import List, Tuple, Optional, Dict

from rf_client.models import Node, NodeType, User
from rf_client.exceptions import *


logger = logging.getLogger('rf_client')

API_BASE_URL = 'http://app.redforester.com/api'
def set_config(api_base_url):
    global API_BASE_URL

    logger.info(f'new api_base_url: {api_base_url}')
    API_BASE_URL = api_base_url


# fixme
class Nodes:
    def __init__(self):
        self._nodes = None

    def _make_tree(self, map_data) -> Node:
        pass

    def get_by_id(self, id: str) -> Node:
        pass

    def get_by_type_id(self, type_id: str) -> Node:
        pass

    def get_all_by_type_id(self, type_id: str):
        pass


class NodeTypes:
    """
    Для болле удобной работы с типами узлов на карте.
    """
    def __init__(self, data):
        self._types = [NodeType(**d) for d in data]

    def find(self, type_name) -> 'NodeType':
        for t in self._types:
            if t.name.lower() == type_name.lower():
                return t

        raise NodeTypeNotFound(f'Can not find {type_name} type in map')


class MapUsers:
    def __init__(self, users):
        self._users = self._data_to_users(users)

    @staticmethod
    def _data_to_users(users_data: List[dict]) -> List[User]:
        users = []
        for ud in users_data:
            users.append(
                User(
                    ud['user_id'],
                    ud['username'],
                    ud['name'],
                    ud['surname'],
                    ud['role']
                ))

        return users

    def find_by_id(self, id_: str) -> User:
        for u in self._users:
            if u.id == id_:
                return u

        raise UserNotFound()

    def find_by_username(self, username) -> User:
        for u in self._users:
            if u.username == username:
                return u

        raise UserNotFound()


def generate_session_id():
    return str(round(random() * 10 ** 10))


async def get_json(response: aiohttp.ClientResponse):
    try:
        return await response.json()
    except (aiohttp.client_exceptions.ClientResponseError, JSONDecodeError, TypeError, ValueError) as e:
        logger.exception('Response was {response.read()}')
        raise RFApiError('Invalid JSON response from RF') from e


class MindMap:
    """
    Класс для работы с деревом карты. Подразумевает работу со снапшотом дерева.
    """
    def __init__(self, map_id: str, credentials: Tuple[str, str], view_root_id=None, session_id=None, timeout: ClientTimeout = None):
        """
        :param map_id: id карты
        :param credentials: пара логин, пароль или другие доступные сочетания
        :param view_root_id: если указан, то загрузка будет происходить от этого узла
        :param session_id: id сесси пользователя, можно указать из вне, но это не требуется
        :param timeout: инстанс aiohttp.ClientTimeout
        """
        logger.debug(f'Init MindMap ({API_BASE_URL}) with map_id={map_id}, view_root_id={view_root_id}, session_id={session_id}')

        self._api_base_url = API_BASE_URL
        self.id = map_id
        self.view_root_id = view_root_id
        self.root: Node = None
        self.ids_to_nodes: Dict[str, Node] = None
        self.nodes: Nodes = None
        self.types: NodeTypes = None
        self.users: MapUsers = None

        username, password = credentials
        if session_id is None:
            session_id = generate_session_id()
            logger.info(f'Generate new session_id: {session_id}')

        self.session_id = session_id

        if timeout is None:
            timeout = ClientTimeout()

        self._session: ClientSession = ClientSession(
            auth=BasicAuth(login=username, password=password),
            json_serialize=ujson.dumps,
            headers={
                'SessionId': self.session_id
            },
            timeout=timeout
        )

    @staticmethod
    def _make_tree(map_data, ignore_out_of_branch=False) -> Tuple[Node, Dict[str, Node]]:
        """
        Соберем дерево из json, которая пришла,
         основная задача метода - правильно обработать узлы ссылки
        :param map_data: json от сервера с GET /nodes
        :param ignore_out_of_branch: иногда полезно проигнорировать ссылки, которые указывают на узлы, которых у нас нет
        :return: Node - корень дерева
        """
        # индекс узлов по их id, чтобы каждый раз не обходить узлы рекурсивно
        ids_to_nodes = dict()
        link_to_postprocess = []

        def dive(node: Node):
            if node.is_link:
                link_to_postprocess.append((node, node.body.id))

            for c in node.body.children:
                dive(c)

        root = Node(**map_data, parent_node=None, ids_to_nodes=ids_to_nodes)

        dive(root)

        for node, link_id in link_to_postprocess:
            linked_node = root.find_by_id(link_id)
            if linked_node:
                node.body = linked_node.body
            else:
                if not ignore_out_of_branch:  # если ссылка вне доступной части карты
                    raise NodeNotFound('Link not found')

        return root, ids_to_nodes


    async def init(self) -> 'MindMap':
        """
        Ленивая инициализация, потом учто конструктор не может быть async
        """
        async with self._session.get('{}/maps/{}/nodes{}'.format(self._api_base_url, self.id, ('/' + self.view_root_id) if self.view_root_id else '')) as response:
            map_data = await get_json(response)

            if response.status == 401:
                raise RFApiError(f"Unauthorized:\n{map_data}")
            elif response.status == 403:
                raise RFApiError(f"Forbidden:\n{map_data}")
            elif response.status >= 400:
                raise RFApiError(f"Other RF API Error:\n{map_data}")

        async with self._session.get('{}/maps/{}/node_types'.format(self._api_base_url, self.id)) as response:
            types_data = await get_json(response)

        async with self._session.get('{}/maps/{}/users'.format(self._api_base_url, self.id)) as response:
            users_data = await get_json(response)

        self.root, self.ids_to_nodes = self._make_tree(map_data, ignore_out_of_branch=self.view_root_id is not None)
        self.types = NodeTypes(types_data)
        self.users = MapUsers(users_data)

        return self

    async def close(self):
        """ Закрыть соединение """
        await self._session.close()

    def __enter__(self) -> 'MindMap':
        """
        Не используется.
        Нужно только для тайпинга, PyCharm проблеммы, не понимает он аннотацию от aenter
        """
        return self

    async def __aenter__(self) -> 'MindMap':
        """
        async with MindMap(<id>, <token>) as mm:
            pass
        """
        await ensure_future(self.init())
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    def get(self, node_id: str) -> Node:
        """
        Получить ссылку на узел по его id.
        """
        try:
            return self.ids_to_nodes[node_id]
        except KeyError:
            raise NodeNotFound()

    async def create(self, parent: Optional[str] = None, parent_node: Optional[Node] = None,
                     node_type: Optional[NodeType] = None, properties: Optional[dict] = None) -> Node:
        """
        Создать узел new_node в RF.
        """
        url = f'{self._api_base_url}/nodes'
        if parent is None and parent_node is None:
            raise RFApiError('Missing parent node')

        node = None
        # todo sessionId miss
        async with self._session.post(
                url,
                json={
                    'parent': parent_node.id if parent_node else parent,
                    'map_id': self.id,
                    'properties': properties,
                    'type_id': node_type.id,
                }
        ) as response:
            resp = await response.read()
            if response.status == 200:
                logger.info('Node has been created')
                # todo put Node to parent in tree
                node = Node(**ujson.loads(resp), ids_to_nodes=self.ids_to_nodes)
            else:
                logger.error(f'Node has NOT been created: {resp}')

        return node

    async def update(self, node: Node, properties_change: Optional[dict] = None) -> Node:
        url = f'{self._api_base_url}/nodes/{node.id}'
        data = {}
        if properties_change:
            data['properties'] = ujson.dumps(properties_change, ensure_ascii=False)

        async with self._session.patch(url, json=data) as response:
            resp = await response.read()
            if response.status == 200:
                node = Node(**ujson.loads(resp), ids_to_nodes=self.ids_to_nodes)
                logger.info(f'Node has been updated: {resp}')
            else:
                logger.error(f'Node has NOT been updated: {resp}')

        return node

    def delete(self, node: Node) -> bool:
        raise NotImplementedError()



class User:
    """
    Модель для пользователя карты
    """
    def __init__(self, id, username, name, surname, role):
        self.id = id
        self.username = username
        self.name = name
        self.surname = surname
        self.role = role

    def __repr__(self):
        return f"User('{self.id}', '{self.username}')"


class RFApiError(Exception):
    pass


class TreeException(Exception):
    """  Базовое исключение  """
    pass


class DataNotFound(TreeException): pass

class NodeNotFound(DataNotFound): pass

class NodeNotInserted(TreeException): pass

class NodeTypeNotFound(DataNotFound): pass

class UserNotFound(DataNotFound): pass



from setuptools import setup, find_packages


with open('README.md') as f:
    long_description = f.read()


setup(name='rf_client',
      version='0.1.1',
      long_description=long_description,
      long_description_content_type='text/markdown',
      description='Python Red Forester async client',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
      ],
      url='https://redforester.com',
      author='Red Forester',
      author_email='tech@redforester.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'aiohttp == 3.4.4',
          'ujson == 1.35',
          'python-dateutil == 2.7.3'
      ],
      include_package_data=True,
      zip_safe=False)



class NodeType:
    """
    Модель типа узла
    """
    class NodeTypeProperty:
        """
        Модель для свойства типа
        """

        def __init__(self, name, default_value, position, **other):
            """
            # todo тут далеко не все атрибуты свойства типа
            :param node_type_property_body:
            """
            self.name = name
            self.position = position
            self.default_value = default_value

        def __str__(self):
            return f"{self.name}: {self.default_value}"

    def __init__(self, id, name, displayable, default_child_node_type_id, icon, properties, **other):
        self.id = id
        self.name = name

        self.displayable = displayable
        self.default_child_node_type_id = default_child_node_type_id
        self.icon = icon
        self.properties = [NodeType.NodeTypeProperty(**t) for t in properties]

    def __repr__(self):
        return f"NodeType('{self.id}', '{self.name}', ...)"